package com.pointrlabs.cordova.pointr;

import android.Manifest;
import android.content.Intent;

import com.pointrlabs.cordova.pointr.core.activity.MapActivity;

import org.apache.cordova.CallbackContext;
import org.apache.cordova.CordovaPlugin;
import org.json.JSONArray;
import org.json.JSONException;

import com.pointrlabs.core.management.PermissionManager;
import com.pointrlabs.core.management.PoiManager;
import com.pointrlabs.core.management.models.Building;
import com.pointrlabs.core.management.models.Level;
import com.pointrlabs.core.management.models.PTRParams;
import com.pointrlabs.core.management.models.PointrEnvironment;
import com.pointrlabs.core.nativecore.wrappers.Plog;
import com.google.gson.Gson;
import com.pointrlabs.core.management.Pointr;
import com.pointrlabs.core.management.PositionManager;
import com.pointrlabs.core.pathfinding.PathFindingMode;
import com.pointrlabs.core.pathfinding.PathManager;
import com.pointrlabs.core.poi.models.Poi;
import com.pointrlabs.core.poi.models.PoiContainer;
import com.pointrlabs.core.positioning.model.CalculatedLocation;

import com.pointrlabs.cordova.pointr.core.models.Location;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.pointrlabs.core.management.DataManager;
import com.pointrlabs.core.management.interfaces.PointrListener;
import com.pointrlabs.core.management.models.DataType;
import com.pointrlabs.core.management.models.ErrorMessage;
import com.pointrlabs.core.management.models.Site;
import com.pointrlabs.core.positioning.model.PositioningTypes;
import com.pointrlabs.core.site.SiteManager;

import java.text.BreakIterator;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;

public class PointrPlugin extends CordovaPlugin {
  ExecutorService backgroundThread = Executors.newCachedThreadPool();
  String licenseKey;
  String environment;
  Integer logLevel = 2;

  PermissionManager permissionManager;

  @Override
  public boolean execute(String action, JSONArray args, CallbackContext callbackContext) {
    switch (action) {
      case "initialize": {
        try {
          licenseKey = args.getString(0);
          environment = args.getString(1);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Critical error! Parameters for initializing Pointr not given properly.");
          callbackContext.error("PointrPlugin ->  Critical error! Could not initialize Pointr Exception: " + ex.getMessage());
          return true;
        }
        try {
          logLevel = args.getInt(2);
        } catch (JSONException ex) {
          Plog.w("PointrPlugin -> Log level not given for initializing Pointr. It will default to VERBOSE");
        }
        callbackContext.success();
        return true;
      }
      case "start": {
        backgroundThread.execute(() -> this.startPointr(callbackContext));
        return true;
      }
      case "stop": {
        backgroundThread.execute(() -> {
          // This is not included in the V1.0
          //this.stopPointr();
        });
        return true;
      }
      case "showPoiDetails": {
        final String siteExternalIdentifier;
        final String poiInternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
          poiInternalIdentifier = args.getString(1);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Could not retrieve the arguments for showing Poi details");
          callbackContext.error("PointrPlugin ->  Could not retrieve the arguments for showing Poi details");
          return true;
        }
        backgroundThread.execute(() -> this.showPoiDetails(siteExternalIdentifier, poiInternalIdentifier, callbackContext));
        return true;
      }
      case "showPathfindingToPoi": {
        final String siteExternalIdentifier;
        final String poiInternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
          poiInternalIdentifier = args.getString(1);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Could not retrieve the arguments for starting path to Poi");
          callbackContext.error("PointrPlugin ->  Could not retrieve the arguments for pathfinding poi details");
          return true;
        }
        backgroundThread.execute(() -> this.showPathfindingToPoi(siteExternalIdentifier, poiInternalIdentifier, callbackContext));
        return true;
      }
      case "showPathfindingBetweenPois": {
        final String siteExternalIdentifier;
        final String sourcePoiInternalIdentifier;
        final String destinationPoiInternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
          sourcePoiInternalIdentifier = args.getString(1);
          destinationPoiInternalIdentifier = args.getString(2);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Could not retrieve siteExternalIdentifier, sourcePoiInternalIdentifier, destinationPoiInternalIdentifier information.");
          callbackContext.error(ex.getMessage());
          return true;
        }
        backgroundThread.execute(() -> this.showPathBetweenPois(siteExternalIdentifier, sourcePoiInternalIdentifier, destinationPoiInternalIdentifier, callbackContext));
        return true;
      }
      case "clearPath": {
        this.clearPath();
        return true;
      }
      case "showBuilding": {
        final String siteExternalIdentifier;
        final String buildingExternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
          buildingExternalIdentifier = args.getString(1);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Could not retrieve the arguments for showing building");
          callbackContext.error(ex.getMessage());
          return true;
        }
        backgroundThread.execute(() -> this.showMapForBuilding(siteExternalIdentifier, buildingExternalIdentifier, callbackContext));
        callbackContext.success();
        return true;
      }
      case "showSite": {
        final String siteExternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Could not retrieve site external identifier. Cannot show site");
          callbackContext.error(ex.getMessage());
          return true;
        }
        backgroundThread.execute(() -> this.showMapForSite(siteExternalIdentifier, callbackContext));
        callbackContext.success();
        return true;
      }
      case "showLevel": {
        final String siteExternalIdentifier;
        final String buildingExternalIdentifier;
        final int levelIndex;
        try {
          siteExternalIdentifier = args.getString(0);
          buildingExternalIdentifier = args.getString(1);
          levelIndex = args.getInt(2);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Could not retrieve building and level index. Cannot show level");
          callbackContext.error(ex.getMessage());
          return true;
        }
        backgroundThread.execute(() -> this.showMapForLevel(siteExternalIdentifier, buildingExternalIdentifier, levelIndex, callbackContext));
        callbackContext.success();
        return true;
      }
      case "showMap": {
        backgroundThread.execute(this::showMap);
        callbackContext.success();
        return true;
      }
      case "highlightPoi": {
        final String siteExternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
        } catch (JSONException e) {
          callbackContext.error(e.getMessage());
          return true;
        }
        final String poiId;
        try {
          poiId = args.getString(1);
        } catch (JSONException e) {
          callbackContext.error(e.getMessage());
          return true;
        }
        if (poiId == null || siteExternalIdentifier == null) {
          Plog.w("PointrPlugin ->  Could not get parameters needed for zooming in to Poi." +
            "Please provide a valid poi id and site external identifier.");
          callbackContext.error("PointrPlugin ->  Wrong parameters for building id and poi id");
          return true;
        }
        Plog.v("PointrPlugin ->  Poi id retrieved: " + poiId);
        Plog.v("PointrPlugin ->  Site id retrieved: " + siteExternalIdentifier);
        backgroundThread.execute(() -> this.highlightPoi(siteExternalIdentifier, poiId, callbackContext));
        return true;
      }
      case "getCurrentLocation": {
        final String typeOfClosestPoi;
        String tempClosestPoi;
        final Integer finalOptionalTimeout;
        Integer optionalTimeout;
        try {
          tempClosestPoi = args.getString(0);
        } catch (JSONException e) {

          tempClosestPoi = null;
        }
        try {
          optionalTimeout = args.getInt(1);
        } catch (JSONException ex) {
          optionalTimeout = null;
        }

        typeOfClosestPoi = tempClosestPoi;
        finalOptionalTimeout = optionalTimeout;
        backgroundThread.execute(() -> this.getCurrentLocation(callbackContext, typeOfClosestPoi, finalOptionalTimeout));
        return true;
      }
      case "showPathfindingToLocation": {
        final String locationJson;
        try {
          locationJson = args.getString(0);
        } catch (JSONException e) {
          callbackContext.error(e.getMessage());
          return true;
        }
        if (locationJson == null) {
          Plog.w("PointrPlugin ->  Location json is not passed. Cannot take user to location.");
          callbackContext.error("PointrPlugin ->  Location json is not passed");
          return true;
        }
        backgroundThread.execute(() -> this.showPathfindingToLocation(locationJson, callbackContext));
        return true;
      }
      case "setPathfindingMode": {
        final int pathfindingMode;
        try {
          pathfindingMode = args.getInt(0);
        } catch (JSONException e) {
          callbackContext.error(e.getMessage());
          return true;
        }
        backgroundThread.execute(() -> this.setPathfindingMode(pathfindingMode));
        return true;
      }
      case "shouldHideRouteSummaryView": {
        final boolean shouldHide;
        try {
          shouldHide = args.getBoolean(0);
        } catch (JSONException e) {
          callbackContext.error(e.getMessage());
          return true;
        }
        MapActivity.Companion.setRouteSummaryHidden(shouldHide);
        return true;
      }
      case "loadDataForSite": {
        final String siteExternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
        } catch (JSONException e) {
          callbackContext.error(e.getMessage());
          return true;
        }
        backgroundThread.execute(() -> this.loadDataForSite(siteExternalIdentifier, callbackContext));
        return true;
      }
    }
    return false;
  }

  private void loadDataForSite(String siteExternalIdentifier, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);
  }

  private void setPathfindingMode(Integer pathfindingMode) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        Pointr pointr = Pointr.getPointr();
        PathManager pathManager = pointr.getPathManager();
        PathFindingMode mode;
        if (pathfindingMode == 0) {
          mode = PathFindingMode.NORMAL;
        } else if (pathfindingMode == 1) {
          mode = PathFindingMode.EASY;
        } else if (pathfindingMode == 2) {
          mode = PathFindingMode.ACCESSIBLE;
        } else {
          mode = PathFindingMode.NORMAL;
        }
        pathManager.setCurrentMode(mode);
      }

      @Override
      public void onPointrFailed(String error) {

      }
    }, false, "");
  }

  private void clearPath() {
    cordova.getActivity().runOnUiThread(() -> {
      Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
      mapActivityIntent.setAction(MapActivity.CLEAR_PATH);
      cordova.getActivity().startActivity(mapActivityIntent);
    });
  }

  private void showPathBetweenPois(String siteExternalIdentifier, String sourcePoiInternalIdentifier, String destinationPoiInternalIdentifier, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        cordova.getActivity().runOnUiThread(() -> {
          Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
          mapActivityIntent.setAction(MapActivity.SHOW_PATHFINDING_BETWEEN_POIS);
          mapActivityIntent.putExtra(MapActivity.site_id_extra_data, siteExternalIdentifier);
          mapActivityIntent.putExtra(MapActivity.source_poi_extra_data, sourcePoiInternalIdentifier);
          mapActivityIntent.putExtra(MapActivity.destination_poi_extra_data, destinationPoiInternalIdentifier);
          cordova.getActivity().startActivity(mapActivityIntent);
        });
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);
  }

  private void initializePointrForEnv(String env, String licenseKey, Integer logLevel) {
    Plog.LogLevel pointrLogLevel;
    PointrEnvironment environment;
    try {
      environment = PointrEnvironment.valueOf(env.toUpperCase());
      pointrLogLevel = Plog.LogLevel.values()[logLevel];
    } catch (IllegalArgumentException ex) {
      Plog.e("PointrPlugin ->  Critical Error!! Pointr is not initialized");
      return;
    }

    PTRParams params = new PTRParams();
    params.setLicenseKey(licenseKey);
    params.setEnvironment(environment);
    params.setLogLevel(pointrLogLevel);
    Pointr.with(cordova.getActivity(), params);
  }


  /**
   * This method ensures;
   * - Pointr is Running
   * - Data management is complete
   */
  private void ensurePointrReady(PointrReadyCallback pointrReadyCallback, Boolean shouldWaitForDataManagement, String siteExternalIdentifier) {
    Pointr pointr = Pointr.getPointr();
    if (pointr == null) {
      initializePointrForEnv(environment, licenseKey, logLevel);
      pointr = Pointr.getPointr();
      if (pointr == null) {
        Plog.e("PointrPlugin ->  Critical Error!! Pointr could not be initialized.");
        pointrReadyCallback.onPointrFailed("PointrPlugin ->  Critical Error!! Pointr could not be initialized.");
        return;
      }
    }
    final Pointr pointrSafe = pointr;

    PointrListener listener = new PointrListener() {
      @Override
      public void onStateUpdated(Pointr.State state) {
        if (state == Pointr.State.RUNNING) {
          ensurePermissionsAreGiven();
          pointrSafe.removeListener(this);
          if (shouldWaitForDataManagement) {
            ensureDataManagementIsComplete(pointrReadyCallback, siteExternalIdentifier);
          } else {
            pointrReadyCallback.onPointrReady();
          }
        } else if (state == Pointr.State.FAILED_REGISTRATION || state == Pointr.State.FAILED_VALIDATION) {
          pointrSafe.removeListener(this);
          pointrReadyCallback.onPointrFailed("PointrPlugin ->  Critical!! Pointr could not be started. All subsequent calls to Pointr methods will fail. Reason: " + state.name());
        }
      }
    };
    pointrSafe.addListener(listener);
    if (pointrSafe.getState() == Pointr.State.RUNNING) {
      ensurePermissionsAreGiven();
      pointrSafe.removeListener(listener);
      if (shouldWaitForDataManagement) {
        ensureDataManagementIsComplete(pointrReadyCallback, siteExternalIdentifier);
      } else {
        pointrReadyCallback.onPointrReady();
      }
    } else {
      pointrSafe.start();
      // Wait for pointrListener
    }
  }

  private void ensurePermissionsAreGiven() {
    permissionManager = Pointr.getPointr().getPermissionManager();
    if (!permissionManager.getHasLocationPermissionWhileInUse()) {
      // request permissions 1 at a time
      cordova.requestPermissions(this, 1, new String[] {Manifest.permission.ACCESS_FINE_LOCATION});
    } else if (!permissionManager.getHasBluetoothPermission()) {
      cordova.requestPermissions(this, 2, new String[] {"android.permission.BLUETOOTH_CONNECT",
        "android.permission.BLUETOOTH_SCAN"});
    }
  }

  @Override
  public void onRequestPermissionResult(int requestCode, String[] permissions, int[] grantResults) throws JSONException {
    super.onRequestPermissionResult(requestCode, permissions, grantResults);
    if (requestCode == 1) {
      if (!permissionManager.getHasBluetoothPermission()) {
        cordova.requestPermissions(this, 2, new String[] {"android.permission.BLUETOOTH_CONNECT",
          "android.permission.BLUETOOTH_SCAN"});
      }
    }
    Plog.v("PointrPlugin ->  All permission requests has been resulted");
  }

  private void ensureDataManagementIsComplete(PointrReadyCallback pointrReadyCallback, String siteExternalId) {
    Pointr pointr = Pointr.getPointr();
    DataManager dataManager = pointr.getDataManager();
    SiteManager siteManager = pointr.getSiteManager();
    Site site = siteManager.getSite(siteExternalId);
    Semaphore semaphore = new Semaphore(1);
    SiteManager.Listener siteManagerListener = semaphore::release;
    siteManager.addListener(siteManagerListener);
    if (site == null) {
      try {
        semaphore.acquire();
        semaphore.tryAcquire(4L, TimeUnit.SECONDS);
        semaphore.release();
      } catch (InterruptedException e) {

      }
      site = siteManager.getSite(siteExternalId);
    }
    if (site == null) {
      Plog.w("PointrPlugin -> Could not load the sites, so cannot load data");
      pointrReadyCallback.onPointrFailed("Could not load the sites");
      return;
    }
    siteManager.removeListener(siteManagerListener);
    final Site requestedSite = site;
   

    DataManager.Listener dataManagerListener = new DataManager.Listener() {
      @Override
      public void onDataManagerStartDataManagementForSite(@NonNull Site site, boolean b) {
        Plog.v("PointrPlugin -> onDataManagerStartDataManagementForSite: " + site.getTitle());
      }

      @Override
      public void onDataManagerBeginProcessingDataForSite(@NonNull Site site, @Nullable DataType dataType, boolean b) {
        String dataTypeName = "";
        if (dataType != null) {
          dataTypeName = dataType.name();
        }
        Plog.v("PointrPlugin -> onDataManagerBeginProcessingDataForSite - data: " + dataTypeName);
      }

      @Override
      public void onDataManagerEndProcessingDataForSite(@NonNull Site site, @Nullable DataType dataType, boolean b, boolean b1, @Nullable List<ErrorMessage> list) {
        String dataTypeName = "";
        if (dataType != null) {
          dataTypeName = dataType.name();
        }
        Plog.v("PointrPlugin -> onDataManagerEndProcessingDataForSite - data: " + dataTypeName);
      }

      @Override
      public void onDataManagerCompleteAllForSite(@NonNull Site site, boolean b, boolean b1, @Nullable List<ErrorMessage> list) {
        Plog.v("PointrPlugin -> onDataManagerCompleteAllForSite site:" + site.getTitle() + " - isOnline: " + b + " - isSuccess: " + b1);
        if (site.hasSameInternalIdentifier(requestedSite) && b && b1) {
          if (dataManager != null) {
            dataManager.removeListener(this);
          }
          pointrReadyCallback.onPointrReady();
        } else if (site.hasSameInternalIdentifier(requestedSite) && b && !b1) {
          if (dataManager != null) {
            dataManager.removeListener(this);
          }
          String errorList = "";
          if (list != null) {
            errorList = Arrays.toString(list.toArray());
          }
          pointrReadyCallback.onPointrFailed(errorList);
        }
      }

      @Override
      public void onDataManagerReadyForSite(@NonNull Site site) {
        Plog.v("PointrPlugin -> onDataManagerReadyForSite");
        if (site.hasSameInternalIdentifier(requestedSite)) {
          // Site is ready
          if (dataManager != null) {
            dataManager.removeListener(this);
          }
          pointrReadyCallback.onPointrReady();
        }
      }
    };
    dataManager.loadDataForSite(requestedSite);
    dataManager.addListener(dataManagerListener);

    if (dataManager.isSiteContentReady(requestedSite)) {
      // Site is ready
      Plog.w("PointrPlugin -> onSiteContentReady for: " + requestedSite.getTitle());
      dataManager.removeListener(dataManagerListener);
      pointrReadyCallback.onPointrReady();
    } else {
      // Wait for dataManager onCompleteAll
    }
  }


  private void startPointr(CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, false, "");

  }


  private void stopPointr() {
    Pointr.getPointr().stop();
  }

  // region Action methods

  private void showPoiDetails(String siteExternalIdentifier, String poiInternalIdentifier, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        openPointrMapActivityWithPoiDetails(siteExternalIdentifier, poiInternalIdentifier);
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);
  }

  private void openPointrMapActivityWithPoiDetails(String siteExternalIdentifier, String poiInternalIdentifier) {
    cordova.getActivity().runOnUiThread(() -> {
      Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
      mapActivityIntent.setAction(MapActivity.SHOW_POI_DETAILS);
      mapActivityIntent.putExtra(MapActivity.site_id_extra_data, siteExternalIdentifier);
      mapActivityIntent.putExtra(MapActivity.destination_poi_extra_data, poiInternalIdentifier);
      cordova.getActivity().startActivity(mapActivityIntent);
    });
  }

  private void showPathfindingToPoi(String siteExternalIdentifier, String poiInternalIdentifier, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        openPointrMapActivityWithPathfindingToPoi(siteExternalIdentifier, poiInternalIdentifier);
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);
  }

  private void openPointrMapActivityWithPathfindingToPoi(String siteExternalIdentifier, String poiInternalIdentifier) {
    cordova.getActivity().runOnUiThread(() -> {
      Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
      mapActivityIntent.setAction(MapActivity.SHOW_PATHFINDING_TO_POI);
      mapActivityIntent.putExtra(MapActivity.site_id_extra_data, siteExternalIdentifier);
      mapActivityIntent.putExtra(MapActivity.destination_poi_extra_data, poiInternalIdentifier);
      cordova.getActivity().startActivity(mapActivityIntent);
    });
  }

  private void showMapForSite(String siteExternalIdentifier, CallbackContext callbackContext) {
    final String siteExtId = siteExternalIdentifier;
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        cordova.getActivity().runOnUiThread(() -> {
          Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
          mapActivityIntent.setAction(MapActivity.SHOW_SITE);
          mapActivityIntent.putExtra(MapActivity.site_id_extra_data, siteExtId);
          cordova.getActivity().startActivity(mapActivityIntent);
        });
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);
  }

  private void showMapForBuilding(String siteExternalIdentifier, String buildingExternalIdentifier, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        cordova.getActivity().runOnUiThread(() -> {
          Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
          mapActivityIntent.setAction(MapActivity.SHOW_BUILDING);
          mapActivityIntent.putExtra(MapActivity.building_id_extra_data, buildingExternalIdentifier);
          cordova.getActivity().startActivity(mapActivityIntent);
        });
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);

  }

  private void showMapForLevel(String siteExternalIdentifier, String buildingExternalIdentifier, Integer levelIndex, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        cordova.getActivity().runOnUiThread(() -> {
          Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
          mapActivityIntent.setAction(MapActivity.SHOW_LEVEL);
          mapActivityIntent.putExtra(MapActivity.building_id_extra_data, buildingExternalIdentifier);
          mapActivityIntent.putExtra(MapActivity.level_id_extra_data, levelIndex);
          cordova.getActivity().startActivity(mapActivityIntent);
        });
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);

  }

  private final int positionWaitTimeoutInSeconds = 60;
  private void getCurrentLocation(CallbackContext callbackContext, String typeOfClosestPoi, Integer timeoutInSeconds) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        Pointr pointr = Pointr.getPointr();
        PositionManager positionManager = pointr.getPositionManager();
        CalculatedLocation currentLocation = positionManager.getCurrentCalculatedLocation();
        if (currentLocation == null || !currentLocation.isValid() || !currentLocation.isGeoValid()) {
          boolean didReceivePosition = false;
          Semaphore positionSemaphore = new Semaphore(1);
          positionManager.addListener(new PositionManager.Listener() {
            @Override
            public void onPositionManagerDetectedPositionLevelChange(@NonNull Level level) {}
            @Override
            public void onPositionManagerPositionIsFading() {}
            @Override
            public void onPositionManagerPositionIsLost() {}
            @Override
            public void onPositionManagerPositioningServiceStateChangedTo(PositioningTypes.PositioningServiceState positioningServiceState) {}
            @Override
            public void onPositionManagerCalculatedLocation(CalculatedLocation calculatedLocation) {
              if (calculatedLocation.isValid() || calculatedLocation.isGeoValid()) {
                positionSemaphore.release();
              }
            }
          });
          try {
            positionSemaphore.acquire();
            int timeout = positionWaitTimeoutInSeconds;
            if (timeoutInSeconds != null) {
              timeout = timeoutInSeconds;
            }
            Plog.i("PointrPlugin -> Position not available yet. Will wait for position manager to heat for: " + timeout + " seconds.");
            didReceivePosition = positionSemaphore.tryAcquire(timeout, TimeUnit.SECONDS);
          } catch (InterruptedException e) { }
          positionSemaphore.release();
          if (!didReceivePosition) {
            Plog.w("PointrPlugin ->  Current location is not valid. Cannot return location.");
            callbackContext.error("PointrPlugin ->  Current user location is not valid. Cannot return location.");
            return;
          } else {
            currentLocation = positionManager.getCurrentCalculatedLocation();
          }
        }
        if (currentLocation.isValid()) {
          if (typeOfClosestPoi != null) {
            final CalculatedLocation finalLocation = currentLocation;
            ensureDataManagementIsComplete(new PointrReadyCallback() {
              @Override
              public void onPointrReady() {
                String poiTitle = "";
                String poiInternalIdentifier = "";
                PoiManager poiManager = pointr.getPoiManager();
                List<String> listOfTypes = new ArrayList<>();
                listOfTypes.add(typeOfClosestPoi);
                PoiContainer poiContainer = poiManager.getPois(finalLocation.getSite(), listOfTypes, true);
                List<Poi> closestPoiList = poiContainer.getPoiList();
                if (!closestPoiList.isEmpty()) {
                  Poi closestPoi = closestPoiList.get(0);
                  poiTitle = closestPoi.getName();
                  poiInternalIdentifier = closestPoi.getId();
                }
                sendLocationJsonToIonic(poiTitle, poiInternalIdentifier, finalLocation, callbackContext);
              }

              @Override
              public void onPointrFailed(String error) {

              }
            }, finalLocation.getSite().getExternalIdentifier());
          } else {
            sendLocationJsonToIonic("","",currentLocation, callbackContext);
          }
        }
        else {
          sendLocationJsonToIonic("","",currentLocation, callbackContext);
        }
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, false, "");

  }

  private void sendLocationJsonToIonic(String poiTitle, String poiInternalIdentifier, CalculatedLocation currentLocation, CallbackContext callbackContext) {
    Building building = currentLocation.getBuilding();
    String buildingExternalIdentifier = "";
    if (building != null) {
      buildingExternalIdentifier = building.getExternalIdentifier();
    }
    Site site = currentLocation.getSite();
    String siteExternalIdentifier = "";
    if (site != null) {
      siteExternalIdentifier = site.getExternalIdentifier();
    }
    Level level = currentLocation.getLevel();
    int levelId = -9999;
    if (level != null) {
      levelId = level.getIndex();
    }
    Location location = new Location(currentLocation.getLat(),
      currentLocation.getLon(),
      levelId,
      buildingExternalIdentifier,
      siteExternalIdentifier,
      poiInternalIdentifier,
      poiTitle
    );
    String locationString = new Gson().toJson(location);
    Plog.v("PointrPlugin -> Sending location back: " + locationString);
    callbackContext.success(locationString);
  }

  private void showPathfindingToLocation(String locationJson, CallbackContext callbackContext) {
    Location location = new Gson().fromJson(locationJson, Location.class);
    if (location == null) {
      callbackContext.error("PointrPlugin ->  Could not get Location");
      return;
    }
    if (location.getSiteExternalIdentifier().isEmpty()) {
      callbackContext.error("PointrPlugin -> Could not get site for location");
      return;
    }
    if (location.getBuildingExternalIdentifier().isEmpty()) {
      callbackContext.error("PointrPlugin -> Could not get building for location");
      return;
    }

    if (location.getPoiInternalIdentifier().isEmpty()) {
      // Means this is an arbitrary location to pathfind to
      ensurePointrReady(new PointrReadyCallback() {
        @Override
        public void onPointrReady() {
          cordova.getActivity().runOnUiThread(() -> {
            Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
            mapActivityIntent.setAction(MapActivity.SHOW_PATHFINDING_TO_LOCATION);
            mapActivityIntent.putExtra(MapActivity.location_extra_data, locationJson);
            cordova.getActivity().startActivity(mapActivityIntent);
          });
          callbackContext.success();
        }

        @Override
        public void onPointrFailed(String error) {
          callbackContext.error(error);
        }
      }, false, "");
    } else {
      // Means we should pathfind to a poi or within a site
      ensurePointrReady(new PointrReadyCallback() {
        @Override
        public void onPointrReady() {
          cordova.getActivity().runOnUiThread(() -> {
            Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
            if (!location.getPoiInternalIdentifier().isEmpty()) {
              mapActivityIntent.setAction(MapActivity.SHOW_PATHFINDING_TO_POI);
              mapActivityIntent.putExtra(MapActivity.destination_poi_extra_data, location.getPoiInternalIdentifier());
              mapActivityIntent.putExtra(MapActivity.site_id_extra_data, location.getSiteExternalIdentifier());
            } else {
              mapActivityIntent.setAction(MapActivity.SHOW_PATHFINDING_TO_LOCATION);
              mapActivityIntent.putExtra(MapActivity.location_extra_data, locationJson);
            }
            cordova.getActivity().startActivity(mapActivityIntent);
          });
          callbackContext.success();
        }

        @Override
        public void onPointrFailed(String error) {
          callbackContext.error(error);
        }
      }, true, location.getSiteExternalIdentifier());
    }


  }

  private void highlightPoi(String siteExternalIdentifier, String poiExternalIdentifier, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        cordova.getActivity().runOnUiThread(() -> {
          Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
          mapActivityIntent.setAction(MapActivity.HIGHLIGHT_POI);
          mapActivityIntent.putExtra(MapActivity.site_id_extra_data, siteExternalIdentifier);
          mapActivityIntent.putExtra(MapActivity.poi_id_extra_data, poiExternalIdentifier);
          cordova.getActivity().startActivity(mapActivityIntent);
        });
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);
  }

  private void showMap() {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
        mapActivityIntent.setAction(MapActivity.SHOW_MAP);
        cordova.getActivity().runOnUiThread(() -> cordova.getActivity().startActivity(mapActivityIntent));
      }

      @Override
      public void onPointrFailed(String error) {

      }
    }, false, "");
  }

  interface PointrReadyCallback {
    void onPointrReady();

    void onPointrFailed(String error);
  }
}
